## procedures for TTS markup

# Task description:
# The task is to add markup tags to an input text using the point-and-click paradigm. 
# The text+tags are stored in the graphical text format, and can be translated into 
# sable format and back again. 
# The user selects text in the text widget. Then a customizer (or shortcuts) allows 
# to attach tags to this text. Some tags don't attach to a highlighted portion of the text 
# but insert a new text (or images) and attach to that. These are called "insertion tags".
# The customizer (popup) should reflect earlier tags which apply to the selected text. 
# Unspecified tags are set to the default in the customizer.
# The user can make multiple modifications to the customizer. Eventually, 
# the tags need to be applied to the text. Conflicts with the existing tags have 
# to be resolved, such as overlapping tags.
# The state of the customizer is contained in the settings of the graphical elements. 
# When the tags need to be applied, we first need to find out which tags have changed, 
# and then apply them to the sable text.

# adding support for a new tag:
# - a tag consists of an element name such as PITCH and at least one attribute name such
#   as LEVEL.
# - specify a default value for each element,attribute pair in proc initializeMarkupGui
# - also specify whether the new tag is of the Insert type
# - define the graphical control elements associated with the new tag in
#   proc buildElements, for example a scale (slider)
# - define the look of the tag in graphtext in proc applyTag, for example 
#   foregroundcolor red
# - define the translation of the new tag into sable in proc graph2sable
#   the attribute values can be mapped using sableValue and graphValue

source markupGui_contour.tcl
source markupGui_lola.tcl
package require Widgets
package require Lyre

# Procedures:

proc openPopup {graphtext name} {
  set highlight [getHighlight $graphtext]
	fakeSelect $graphtext create $highlight
  set tags [findSurroundingTags $graphtext $highlight]
  setElementValues $tags  
  rememberElementValues
  buildElements $graphtext $highlight $name
}

proc playPopup {graphtext highlight w} {
  set tmpGraphtext [copyTextWidget $graphtext $graphtext.copy]
  applyPopup $tmpGraphtext $highlight $w
  $tmpGraphtext delete [lindex $highlight 1] end
  $tmpGraphtext delete 1.0 [lindex $highlight 0]
  playGraphtext $tmpGraphtext
  destroy $tmpGraphtext
}

proc playGraphtext {graphtext} {
  set sabletext [copyTextWidget $graphtext $graphtext.sable]
  graph2sable $sabletext
  mkdoTTS [$sabletext get 0.0 end]
  destroy $sabletext 
}

proc cancelPopup {graphtext w} {
  global MarkupGui
	
	fakeSelect $graphtext destroy
	destroy $w
}
	
proc okPopup {graphtext highlight w} {
  global MarkupGui
	
	fakeSelect $graphtext destroy
	applyPopup $graphtext $highlight $w
	destroy $w
}
	
proc applyPopup {graphtext highlight w} {
  set tags [findChangedElements $w $highlight]
  foreach newTag $tags {
    set otherTags [findSableTags $graphtext] 
    # note: findSurroudingTags only finds tags that fully embed a given tag
    # here we need all potentially overlapping tags, so the request routine
    # can split the new tag up if necessary
    requestTag $newTag $otherTags $graphtext $highlight
  }
  if [winfo ismapped $graphtext] {
    propagateFontChanges $graphtext
  }
}  

proc getHighlight {graphtext} {
	global MarkupGui
	
  set highlight [$graphtext tag ranges sel]
  if [expr [llength $highlight] == 0] {
    set indx [$graphtext index insert]
    set highlight "$indx $indx"
  } else {
		# trim spaces
		foreach {indx1 indx2} $highlight break
		while {[regexp " " [$graphtext get $indx1]] & 
           [$graphtext compare $indx1 < $indx2]} {
			set indx1 [$graphtext index $indx1+1c]
		}
		while {[regexp " " [$graphtext get $indx2-1c]] &
           [$graphtext compare $indx1 < $indx2]} {
			set indx2 [$graphtext index $indx2-1c]
		}
		set highlight "$indx1 $indx2"
	}
  return $highlight
}

# find all the tags that apply to the entire highlighted region in graphtext
# since the tags are unique and have only 1 range, it is sufficient to find the tags that 
# apply to both the first and last index in highlight
proc findSurroundingTags {graphtext highlight} {
  foreach {hlBegin hlEnd} $highlight break 
  set tagsBegin [findSableTags $graphtext $hlBegin]
  set tagsEnd [findSableTags $graphtext $hlEnd-1c]
  set tags [intersect $tagsBegin $tagsEnd]
  return $tags
}

proc findSurrInsertTag {graphtext highlight} {
  global MarkupGui
	
	set surrInsertTags ""
	set otherTags [findSurroundingTags $graphtext $highlight]
	foreach otherTag $otherTags {
		foreach {element ID} [split $otherTag _] break
		if $MarkupGui($element,isInsert) {
			lappend surrInsertTags $otherTag
		}
	}
	set len [llength $surrInsertTags]
	if [expr $len > 1] {
		puts "warning in findSurrInsertTags: found $len > 1 Insert tag around $highlight"
	}
	return [lindex $surrInsertTags 0]
}
			
# open a notebook gui with the elements set to the correct positions
proc buildElements {graphtext highlight w} {
  global MarkupGui

  # setup toplevel. 
  if {[winfo exists $w]} {catch {destroy $w}}
  toplevel $w
  #bind $graphtext <Destroy> "catch {destroy $w}"
  wm minsize $w 440 310
  set title [eval "$graphtext get $highlight"]
  if {$title == ""} { set title "insert markup tag" }
  wm title $w $title
  wm geometry $w +440+400
  set b $w.book
  notebook $b
  pack $b -expand 1 -fill both -padx 5 -pady 5 
  set buttonWidth 8
 	set bottomFrame [frame $w.bottom]; pack $bottomFrame -side bottom
	button $bottomFrame.apply -command "applyPopup $graphtext {$highlight} $w" -text Apply -padx 5 -width $buttonWidth 
	button $bottomFrame.ok -command "okPopup $graphtext {$highlight} $w" -text Ok -padx 5 -width $buttonWidth
	button $bottomFrame.cancel -command "cancelPopup $graphtext $w" -text Cancel -padx 5 -width $buttonWidth
  bind $w <Key-Return> "okPopup $graphtext {$highlight} $w"
  pack $bottomFrame.ok -side right
  pack $bottomFrame.apply -side right
  pack $bottomFrame.cancel -side left
  
  #
  # build all elements of the markup popup
  #
  
	foreach {indx1 indx2} $highlight break
	if [$graphtext compare $indx1 != $indx2] {
    #
		# Emphasis page
    #
		set page [$b addpage "emphasis" -text "Emphasis"]; pack $page
		set rightFrame [frame $page.frame2]
		set labelWidth 10
		set emphElements ""
    foreach element {PITCH,BASE PITCH,RANGE RATE,SPEED VOLUME,LEVEL} {
			set elementPath [string tolower $element]
			pack [frame $rightFrame.$elementPath] -fill x
			pack [label $rightFrame.$elementPath.l -text $elementPath -width $labelWidth -anchor w] -side left
			pack [label $rightFrame.$elementPath.l2 -textvariable MarkupGui(popup,$element) \
						-width 5 -anchor e -padx 3 -foreground red] -side left
      pack [label $rightFrame.$elementPath.l3 -text "%" -fg red] -side left
			scale $rightFrame.$elementPath.s -from 30 -to 300 \
				-resolution 1 -orient horizontal \
				-length 6c -width .3c -tickinterval 0 -showvalue no \
				-variable MarkupGui(popup,$element)
      lappend emphElements $rightFrame.$elementPath.s
			pack $rightFrame.$elementPath.s -side left -pady 3
		}
		set element PITCH,TOBI
		set elementPath [string tolower $element]
		pack [frame $rightFrame.$elementPath] -fill x
		pack [label $rightFrame.$elementPath.l -text $elementPath -width $labelWidth -anchor w] -side left
		pack [label $rightFrame.$elementPath.l2 -textvariable MarkupGui(popup,$element) \
					-width 5 -anchor e -padx 3 -foreground red] -side left
    pack [label $rightFrame.$elementPath.l3 -text "   " -fg red] -side left
		entry $rightFrame.$elementPath.t1 -width 7 -textvariable MarkupGui(popup,$element)
    lappend emphElements $rightFrame.$elementPath.t1
    pack $rightFrame.$elementPath.t1 -padx 2 -side left 
		set leftFrame [frame $page.frame1] 
    pack $leftFrame $rightFrame -side top -anchor nw -pady 5 -padx 10
		pack [frame $leftFrame.presets] -fill x
    createEmphListbox $leftFrame.presets.lb $leftFrame.presets $emphElements
    #pack $leftFrame.presets.l -side left -anchor nw
		
    # configure labels and StickyNotes
		# StickyNote attach $rightFrame.pitchB.s "lower  <----- | ----->  higher"
		# since this is the user friendly part I'll leave it for later

    #
    # Sayas page
    #
    set page [$b addpage "sayas" -text "Say As"]; pack $page 
    set page [frame $page.f]; pack $page -anchor nw -fill both
    set leftFrame [frame $page.frame1]; 
    set rightFrame [frame $page.frame2 -borderwidth 1 -relief solid]; 
    pack $leftFrame $rightFrame -side left -anchor nw -pady 20 -padx 10
    pack configure $rightFrame -expand 1 -fill both
    listbox $leftFrame.lb -borderwidth 1 -height 6 -width 13 -bg white -exportselection 0
    pack $leftFrame.lb -side left 
    set elem [sayasFindCurrentElement]
    #if {$elem == "Normal"} {set elem Pronounce}
    foreach name $MarkupGui(sayasCommands) {
      set name [lindex $name 0]
      $leftFrame.lb insert end $name
      if {$name == $elem} {
        $leftFrame.lb selection set end
      }
    }
    sayasCreateElement $rightFrame $elem
    focus $leftFrame.lb
    bind $leftFrame.lb <Button-1> "sayasUpdateCmd $leftFrame.lb $rightFrame"
    bind $leftFrame.lb <Key> "sayasUpdateCmd $leftFrame.lb $rightFrame"
    set bt [bindtags $leftFrame.lb]
    bindtags $leftFrame.lb "[lrange $bt 1 end] [lindex $bt 0]"
    
    if $MarkupGui(detailsEnabled) {
      #
      # Details page
      #
      set page [$b addpage "details" -text "Details"]; pack $page
      $b.emphasis select ;# do this here while details page is still being worked on
      frame $page.f0contour; frame $page.bottom
      foreach {segs contour} [findDetails $graphtext $highlight] break
      if {[llength $segs] > 0} {
        set dur [expr [lindex [lindex $segs end] 1]]
      } else { set dur 1000 }
      llola $page.segs $dur -height 30
      configureLabelWidget $page.segs
      foreach {defaultSegs defaultContour} [defaultDetails $graphtext $highlight $w] break
      button $page.bottom.restore -text "Default" -pady 0 -command "eraseContour; \
        showContour {$defaultContour} $MarkupGui(highlightBeginTime) $MarkupGui(highlightEndTime);\
        $page.segs load {$defaultSegs}"
      pack $page.bottom.restore -padx 5 -pady 3 -side left -expand 0
      initContour $page.f0contour $page.bottom 
      showContour $contour $MarkupGui(highlightBeginTime) $MarkupGui(highlightEndTime)
      set MarkupGui(remember,PITCH,CONTOUR) [normalizeContour $contour]
      set MarkupGui(remember,PRON,SEGDUR) [normalizeSegs $segs]
      # pass default contour and segs through GUIs to know quantization errors
      set defaultContour [conversion [conversion $defaultContour 0] 1]
      $page.segs load $defaultSegs
      set defaultSegs [$page.segs get]
      set MarkupGui(default,PITCH,CONTOUR) [normalizeContour $defaultContour]
      set MarkupGui(default,PRON,SEGDUR) [normalizeSegs $defaultSegs]
      $page.segs load $segs
      pack $page.f0contour $page.segs $page.bottom -side top -fill x
      pack configure $page.f0contour -fill both -expand 1
      $page.segs configure -fittowindow 1
    } else {
      # default selected page
      $b.emphasis select
    }
	} else {
		# 
		# break insert page
		#
		set page [$b addpage "break" -text "Break"]; pack $page
		set leftFrame [frame $page.frame1]; 
    pack $leftFrame -side left -anchor nw -pady 20 -padx 10
		set rightFrame [frame $page.frame2]; 
    pack $rightFrame -fill y -side right -anchor n -pady 15 -padx 5
		set labelWidth 10
    listbox $leftFrame.lb -borderwidth 2  -height 6 -width 13 -background white
    pack $leftFrame.lb -side left -pady 5
    pack [label $rightFrame.dummy -height 5] -side top
    pack [frame $rightFrame.msec] -fill x -side top
    pack [label $rightFrame.msec.l -text Duration -padx 2 -anchor w] -side left
    pack [label $rightFrame.msec.l2 -textvariable MarkupGui(popup,BREAK,MSEC) \
          -width 5 -anchor e -padx 3 -foreground red] -side left
    pack [label $rightFrame.msec.l3 -text " msec" -fg red] -side left
    scale $rightFrame.msec.s -from 0 -to 500 \
      -resolution 10 -orient horizontal \
      -length 6c -width .3c -tickinterval 0 -showvalue no \
      -variable MarkupGui(popup,BREAK,MSEC)
    pack $rightFrame.msec.s -side left -pady 3
		foreach breakType $MarkupGui(breakTypes) {
			$leftFrame.lb insert end [lindex $breakType 0] }
    if {$MarkupGui(popup,BREAK,LEVEL) != ""} {
      $leftFrame.lb selection set $MarkupGui(popup,BREAK,LEVEL) 
    }
    if {$MarkupGui(popup,BREAK,MSEC) != $MarkupGui(default,BREAK,MSEC)} {
      $leftFrame.lb selection clear 0 end
      $leftFrame.lb selection set 4 
    }
    breakUpdateCmd $leftFrame.lb $rightFrame.msec.s
		bind $leftFrame.lb <Button-1> "breakUpdateCmd $leftFrame.lb $rightFrame.msec.s"
		bind $leftFrame.lb <Key> "breakUpdateCmd $leftFrame.lb $rightFrame.msec.s"
		set bt [bindtags $leftFrame.lb]
		bindtags $leftFrame.lb "[lrange $bt 1 end] [lindex $bt 0]"
    set cmd "$leftFrame.lb selection clear 0 end;$leftFrame.lb selection set 4;\
             breakUpdateCmd $leftFrame.lb $rightFrame.msec.s"
    bind $rightFrame.msec.s <ButtonRelease-1> $cmd    
    bind $rightFrame.msec.s <Key> $cmd  
    focus $leftFrame.lb
    
    #
    # Speakers page
    #
    set page [$b addpage "speakers" -text "Speakers"]; pack $page
    set leftFrame [frame $page.frame1]; 
    pack $leftFrame -side left -anchor nw -pady 20 -padx 10
    set rightFrame [frame $page.frame2 -borderwidth 1]; 
    pack $rightFrame -side left -anchor n -pady 15 -padx 5 
    listbox $leftFrame.lb -borderwidth 2  -height 6 -width 25 -bg white
    pack $leftFrame.lb -side left -pady 5
    if [info exists MarkupGui(ttsCmd)] {
      set ttsCmd $MarkupGui(ttsCmd)
      set voices [$ttsCmd getVoiceList]
      set lines ""
      foreach voice $voices {
        array set tmpArray [$ttsCmd voiceInfo $voice]
        lappend lines "$voice - $tmpArray(dialect) $tmpArray(language) $tmpArray(gender)"
      }
    } else {
      set lines "{mwm - American English male}"
    }
    foreach line $lines {
      set voice [lindex $line 0]
      $leftFrame.lb insert end $line
      if {$voice == $MarkupGui(popup,SPEAKER,NAME)} {
        $leftFrame.lb selection set end
      }
    }
    bind $leftFrame.lb <Button-1> "listUpdateCmd $leftFrame.lb SPEAKER,NAME"
    bind $leftFrame.lb <Key> "listUpdateCmd $leftFrame.lb SPEAKER,NAME"
    set bt [bindtags $leftFrame.lb]
    bindtags $leftFrame.lb "[lrange $bt 1 end] [lindex $bt 0]"
    
    #
    # Sounds page
    #
    set page [$b addpage "sounds" -text "Sounds"]; pack $page
    set leftFrame [frame $page.frame1]; 
    pack $leftFrame -side left -anchor nw -pady 20 -padx 10
    set rightFrame [frame $page.frame2 -borderwidth 1]; 
    pack $rightFrame -side left -anchor n -pady 15 -padx 5
    listbox $leftFrame.lb -borderwidth 2  -height 6 -width 15 -bg white \
      -yscrollcommand "$leftFrame.s set"
    scrollbar $leftFrame.s -orient vertical -command "$leftFrame.lb yview"
    grid $leftFrame.lb $leftFrame.s -sticky news
    grid columnconfigure $leftFrame 0 -weight 1
    grid rowconfigure $leftFrame 0 -weight 1
    #pack $leftFrame.lb -side left -pady 5
    if [info exists MarkupGui(sounds)] {
      eval "$leftFrame.lb insert end $MarkupGui(sounds)"
      bind $leftFrame.lb <Button-1> "listUpdateCmd $leftFrame.lb AUDIO,SRC"
      bind $leftFrame.lb <Key> "listUpdateCmd $leftFrame.lb AUDIO,SRC"
      set bt [bindtags $leftFrame.lb]
      bindtags $leftFrame.lb "[lrange $bt 1 end] [lindex $bt 0]"
    } else {
      $leftFrame.lb insert end "(no sounds defined)"
    }

    #
    # Baldi page
    #
    set page [$b addpage "baldi" -text "Baldi"]; pack $page
    set emoFrame [titledFrame $page.f1 -text Emotions]
    set gestFrame [titledFrame $page.f2 -text Gestures]
    pack $page.f1 $page.f2 -side top -fill both -expand 1
    foreach emo $MarkupGui(baldiEmotions) {
      set emo [lindex $emo 0]
      set imageNum [lsearch $MarkupGui(images) $emo]
      button $emoFrame.$emo -command " \
        set MarkupGui(popup,MARKER,EMBED) emo:$emo; \
        $emoFrame.$emo configure -relief sunken"
      if {$imageNum > -1} {
        set im [lindex $MarkupGui(images) $imageNum] 
        $emoFrame.$emo configure -image $im
      } else {
        $emoFrame.$emo configure -text $emo
      }
      pack $emoFrame.$emo -side left
    }
    foreach gest $MarkupGui(baldiGestures) {
      set imageNum [lsearch $MarkupGui(images) $gest]
      button $gestFrame.$gest -command " \
        set MarkupGui(popup,MARKER,EMBED) gest:$gest; \
        $gestFrame.$gest configure -relief sunken"
      if {$imageNum > -1} {
        set im [lindex $MarkupGui(images) $imageNum] 
        $gestFrame.$gest configure -image $im
      } else {
        $gestFrame.$gest configure -text $gest
      }
      pack $gestFrame.$gest -side left
    }
   
    # default selected page
    $b.break select
  }
}

# generic list update command. Problem is that every list is special.
proc listUpdateCmd {lb field} {
  global MarkupGui
  
  set indx [$lb curselection]
  if {$indx == ""} return ;# can this happen?
  set sel [$lb get $indx]
  set MarkupGui(popup,$field) [lindex $sel 0]
}

proc breakUpdateCmd {lb msec} {
	global MarkupGui
	
	set indx [$lb curselection]
  if {$indx < 4} {
    set MarkupGui(popup,BREAK,LEVEL) $indx
    set MarkupGui(popup,BREAK,MSEC) $MarkupGui(remember,BREAK,MSEC)
    $msec set 0
    [winfo parent $msec].l configure -foreground "#666"
    [winfo parent $msec].l2 configure -foreground "#666"
    [winfo parent $msec].l3 configure -foreground "#666"
    focus $lb
  } elseif {$indx == 4} {
    set MarkupGui(popup,BREAK,LEVEL) $MarkupGui(remember,BREAK,LEVEL)
    #set MarkupGui(popup,BREAK,MSEC) 100
    [winfo parent $msec].l configure -foreground black
    [winfo parent $msec].l2 configure -foreground red
    [winfo parent $msec].l3 configure -foreground red
    focus $msec
  }
}
		
# Return listbox which handles emphasis presets
# When the selection changes, the MarkupGui(popup,X) values will be updated 
# according to the presets. The selection is stored in MarkupGui(popup,emphListbox)
# This procedure should also listen for changes to the X variables
proc createEmphListbox {listBox leftFrame elements} {
  global MarkupGui
  
  label $leftFrame.l -text "Emphasis Presets           "
  listbox $listBox -borderwidth 2  -height 4 -width 13 -bg white \
      -yscrollcommand "$leftFrame.s set" -exportselection 0
  scrollbar $leftFrame.s -orient vertical -command "$listBox yview"
  button $leftFrame.add -text "Add" -command "emphAddPreset $listBox" -pady 0
  button $leftFrame.remove -text "Remove" -command "emphRemovePreset $listBox" -pady 0
  grid $leftFrame.l $leftFrame.lb $leftFrame.s $leftFrame.add $leftFrame.remove -sticky news
  grid configure $leftFrame.l -sticky nw
  grid configure $leftFrame.add $leftFrame.remove -sticky sw -padx 3
  grid columnconfigure $leftFrame 0 -weight 1
  grid rowconfigure $leftFrame 0 -weight 1
  foreach preset $MarkupGui(emphasisPresets) {
    $listBox insert end [lindex $preset 0]
  }
  # see if the current settings correspond to a preset
  emphCheckPreset $listBox
  bind $listBox <Button-1> "emphUpdateCmd %W"
  bind $listBox <Key> "emphUpdateCmd %W"
  set bt [bindtags $listBox]
  bindtags $listBox "[lrange $bt 1 end] [lindex $bt 0]"
  foreach elem $elements {
    bind $elem <ButtonRelease-1> "emphCheckPreset $listBox" 
    bind $elem <Key> "emphCheckPreset $listBox"
  }
  return $listBox
}

proc emphAddPreset {w} {
  global MarkupGui
  
  # get a name
  set t [toplevel $w.presets]
  wm title $t "Add Preset"
  label $t.l -text "Add the current settings as a Preset."
  frame $t.name ; frame $t.buttons
  label $t.name.l2 -text "Name:"
  set num 1
  set currentNames [$w get 0 end]
  while {[lsearch $currentNames preset$num] > -1} {incr num}
  entry $t.name.e
  $t.name.e insert 0 "preset$num"
  button $t.buttons.ok -text OK -width 12 -command "lappend MarkupGui(emphasisPresets) \
    {[list [$t.name.e get] PITCH,BASE $MarkupGui(popup,PITCH,BASE) \
    PITCH,RANGE $MarkupGui(popup,PITCH,RANGE) PITCH,TOBI $MarkupGui(popup,PITCH,TOBI) \
    VOLUME,LEVEL $MarkupGui(popup,VOLUME,LEVEL) RATE,SPEED $MarkupGui(popup,RATE,SPEED)]} ;\
    saveSettings; $w insert end [$t.name.e get]; $w selection clear 0 end; \
    $w selection set end;destroy $t"
  button $t.buttons.cancel -text Cancel -width 12 -command "destroy $t"
  
  pack $t.l $t.name $t.buttons -side top -pady 20
  pack $t.name.l2 $t.name.e -side left -padx 5
  pack $t.buttons.ok $t.buttons.cancel -side left -padx 20
}

proc emphRemovePreset {w} {
  global MarkupGui
  
  set indx [$w curselection]
  if {$indx != ""} {
    $w delete $indx
    set MarkupGui(emphasisPresets) [lreplace $MarkupGui(emphasisPresets) $indx $indx]
    saveSettings
  }
}
  
proc emphUpdateCmd {w} {
  global MarkupGui
  
  set currentPreset [$w curselection]
  if {$currentPreset != ""} {
    set preset [$w get $currentPreset]
    foreach key {PITCH,BASE PITCH,RANGE VOLUME,LEVEL RATE,SPEED PITCH,TOBI} {
      set MarkupGui(popup,$key) $MarkupGui(default,$key)
    }
    set presets [$w get 0 end]
    set indx [lsearch $presets $preset] ;# should always exist
    set settings [lrange [lindex $MarkupGui(emphasisPresets) $indx] 1 end]
    foreach {key value} $settings {
      set MarkupGui(popup,$key) $value
    }
  }
}

# check if current settings of pitch sliders etc. correspond with a preset.
# show correct selection on listbox
proc emphCheckPreset {w} {
  global MarkupGui
  
  set len [llength $MarkupGui(emphasisPresets)]
  for {set i 0} {$i < $len} {incr i} {
    set preset [lindex $MarkupGui(emphasisPresets) $i]
    array set tmpArray [lrange $preset 1 end]
    set b 1
    foreach key [array names tmpArray] {
      set b [expr $b & $MarkupGui(popup,$key) == $tmpArray($key)]
    }
    if $b { 
      $w selection clear 0 end
      $w selection set $i
      return
    }
  }
  $w selection clear 0 end
}
 
# handle behavior of the sayas listbox
# build rightFrame according to listbox selection
# set correct fields in global variable MarkupGui 
proc sayasUpdateCmd {lb rf} {
  global MarkupGui
  
  set element ""
  # current sayas fields: SAYAS,MODE,MODETYPE and PRON,SUB|IPA
  # sayas commands cannot be combined. Give up previous selection?
  if {$MarkupGui(popup,SAYAS,MODE) != $MarkupGui(default,SAYAS,MODE) | \
      $MarkupGui(popup,SAYAS,MODETYPE) != $MarkupGui(default,SAYAS,MODETYPE)} {
    set element SAYAS; set atts "MODE: $MarkupGui(popup,SAYAS,MODE) \n"
    append atts "MODETYPE: $MarkupGui(popup,SAYAS,MODETYPE)"
  } elseif {$MarkupGui(popup,PRON,SUB) != $MarkupGui(default,PRON,SUB)} { 
    set element PRON; set atts "SUB: $MarkupGui(popup,PRON,SUB)"
  } elseif {$MarkupGui(popup,PRON,IPA) != $MarkupGui(default,PRON,IPA)} {
    set element PRON; set atts "IPA: $MarkupGui(popup,PRON,IPA)"
  }
  if {$element != ""} {
    set msg "You have currently selected:\n$element\n$atts\n\n"
    append msg "Sayas commands cannot be combined.\n\n"
    append msg "Do you want to erase the previous command and create a new one ?" 
    # set ans [tk_messageBox -type yesno -message $msg]
    # JW: disabled this because jacques don't like dialog boxes
    set ans "yes"
    focus $rf
    if {$ans == "no"} {
      set elem [sayasFindCurrentElement]
      $lb selection clear 0 end
      $lb selection set [lsearch [$lb get 0 end] "$elem"]
      return
    } else {
      set MarkupGui(popup,SAYAS,MODE) $MarkupGui(default,SAYAS,MODE)
      set MarkupGui(popup,SAYAS,MODETYPE) $MarkupGui(default,SAYAS,MODETYPE)
      set MarkupGui(popup,PRON,SUB) $MarkupGui(default,PRON,SUB)
      set MarkupGui(popup,PRON,IPA) $MarkupGui(default,PRON,IPA)
    }
  }
  # create rightFrame corresponding to selection
  set indx [$lb curselection]
  if {$indx == ""} {
    # not sure how this could happen since you can't delete all selections using the mouse
    return 
  }
  set key [$lb get $indx]
  regsub " " $key "" key
  sayasCreateElement $rf $key
  focus $lb
}

proc sayasCreateElement {rf key} {
  global MarkupGui
  
  # start with clean rightFrame
  foreach elem [winfo children $rf] {
    destroy $elem
  }
  set labelWidth 15
  set key [string tolower $key]
  switch $key {
    pronounce {
      pack [frame $rf.sub] -side top -anchor nw -padx 5 -pady 5
      pack [label $rf.sub.l -text "Substitute word(s)" -width $labelWidth] -side left
      entry $rf.sub.e -width $labelWidth -textvariable MarkupGui(popup,PRON,SUB)
      pack $rf.sub.e -side left
      pack [frame $rf.worldbet] -side top -anchor nw -padx 5 -pady 5
      pack [label $rf.worldbet.l -text "Worldbet symbols" -width $labelWidth] -side left
      entry $rf.worldbet.e -width $labelWidth -textvariable MarkupGui(popup,PRON,IPA)
      pack $rf.worldbet.e -side left
    }
    digits -
    spell {
      set MarkupGui(popup,SAYAS,MODE) literal
      if {$MarkupGui(popup,SAYAS,MODETYPE) == $MarkupGui(default,SAYAS,MODETYPE)} {
        set MarkupGui(popup,SAYAS,MODETYPE) fluent
      }
      radiobutton $rf.cb -text fluent -variable MarkupGui(popup,SAYAS,MODETYPE) \
        -value fluent
      radiobutton $rf.ncb -text slow -variable MarkupGui(popup,SAYAS,MODETYPE) \
        -value isolated
      pack $rf.cb -side top -anchor nw
      pack $rf.ncb -anchor nw
    }
    syllabify {
      set MarkupGui(popup,SAYAS,MODE) syllabify
      frame $rf.hand
      radiobutton $rf.cb -text automatic -variable cb -value 0 -command {
        set MarkupGui(popup,SAYAS,MODETYPE) ""
      }
      radiobutton $rf.hand.ncb -text phonemes -variable cb -value 1
      entry $rf.hand.e -width 13 -textvariable MarkupGui(popup,SAYAS,MODETYPE) 
      bind $rf.hand.e <Button-1> "set cb 1"
      bind $rf.hand.e <Key> "set cb 1"
      pack $rf.cb -side top -anchor nw
      pack $rf.hand -side top -anchor nw
      pack $rf.hand.ncb -side left
      pack $rf.hand.e -side left -padx 5
      set cb ![string compare $MarkupGui(popup,SAYAS,MODETYPE) ""]
    }
    amount {
      set MarkupGui(popup,SAYAS,MODE) cardinal
    }
    normal -
    articulation -
    default {
      # do nothing
    }
  }
}
      
proc sayasFindCurrentElement {} {
  global MarkupGui
  
  if {$MarkupGui(popup,SAYAS,MODE) == "syllabify"} {return Syllabify}
  if {$MarkupGui(popup,SAYAS,MODE) == "literal"} {
    if [regexp {\+?-?[0-9][0-9.,]*} $MarkupGui(popup,SAYAS,MODETYPE)] {
      return Digits
    } else {
      return Spell
    }
  }
  if {$MarkupGui(popup,PRON,SUB) != "" | \
      $MarkupGui(popup,PRON,IPA) != ""} {return Pronounce}
  return Normal
}
  
# set MarkupGui(popup,elementName) to either default or value determined by tag
proc setElementValues {tags} {
  global MarkupGui
  
  # tags correspond to tag names of a textwidget somewhere
  # these are unique names that need to be decoded here into element names
  # we also keep a data structure somewhere with the attribute values for a tag
  foreach element $MarkupGui(elementNames) {
    foreach att $MarkupGui($element,attNames) {
      set MarkupGui(popup,$element,$att) $MarkupGui(default,$element,$att)
    }
  }
  foreach tag $tags {
    set parseTag [split $tag _]
    foreach {element ID} $parseTag break
    foreach att $MarkupGui($element,attNames) {
      if [info exists MarkupGui(attributes,$tag,$att)] {
        set MarkupGui(popup,$element,$att) $MarkupGui(attributes,$tag,$att)
      }
    }
  }
}

proc rememberElementValues {} {
  global MarkupGui
  
  foreach element $MarkupGui(elementNames) {
    foreach att $MarkupGui($element,attNames) {
      set MarkupGui(remember,$element,$att) $MarkupGui(popup,$element,$att)
    }
  }
}

# since there is no way to listen for the deletion of a tag in a text widget, 
# this function should be called from time to time to delete attributes of a
# tag that has been deleted in the main text widget. Note that at several occasions 
# a copy can be made of the main text widget (for example to audio-test it), and tags 
# can be deleted in the copy, but remain alive in the main text widget.
proc cleanupTags {graphtext} {
  global MarkupGui
  
	set names [array names MarkupGui "attributes,*,*"]
	foreach name $names {
		set name [lindex [split $name ,] 1] 
		set range [$graphtext tag ranges $name]
		if [expr [llength $range] == 0] {
			cleanupAtts $name
		}
	}
}

proc cleanupAtts {tagName} {
  global MarkupGui
  
  if {$tagName == "all"} {
    set names [array names MarkupGui "attributes,*_*,*"]
    eval "lappend names [array names MarkupGui popup,*_*,*]"
    eval "lappend names [array names MarkupGui remember,*_*,*]"
  } else {
    set names [array names MarkupGui "*,$tagName,*"]
  }
  foreach name $names {
    unset MarkupGui($name)
  }
}

# find which current popup values MarkupGui(popup,X) are different from
# remembered values MarkupGui(remember,X)
# For the "details" page the popup values don't get automatically updated 
# so we have to query these
proc findChangedElements {w highlight} {
  global MarkupGui
  
  foreach {indx1 indx2} $highlight break
  if {$MarkupGui(detailsEnabled) & $indx1 != "" & $indx2 != "" & $indx1 != $indx2} {
	  # query pitch contour and label widget
	  set contour [normalizeContour [extractContour]]
	  set MarkupGui(popup,PITCH,CONTOUR) $contour
	  if [winfo exists $w.book.details.segs] {
      set MarkupGui(popup,PRON,SEGDUR) [normalizeSegs [$w.book.details.segs get]] }
    #puts "DB: default contour: $MarkupGui(default,PITCH,CONTOUR)"
    #puts "DB: popup contour: $MarkupGui(popup,PITCH,CONTOUR)"
    #puts "DB: default segs: $MarkupGui(default,PRON,SEGDUR)"
    #puts "DB: popup segs: $MarkupGui(popup,PRON,SEGDUR)"
	} else {
    set MarkupGui(popup,PITCH,CONTOUR) $MarkupGui(default,PITCH,CONTOUR)
    set MarkupGui(popup,PRON,SEGDUR) $MarkupGui(default,PRON,SEGDUR)
  }
  
  set elements ""
  foreach element $MarkupGui(elementNames) {
    set elementChanged 0
    foreach attName $MarkupGui($element,attNames) {
			if ![string match $MarkupGui(popup,$element,$attName) \
                        $MarkupGui(remember,$element,$attName)] {
        set elementChanged 1
      }
    }
    set atts ""
    set attNames ""
    if $elementChanged {
      foreach attName $MarkupGui($element,attNames) {
        if ![string match $MarkupGui(popup,$element,$attName) \
                          $MarkupGui(default,$element,$attName)] {
          lappend attNames $attName
          lappend atts $MarkupGui(popup,$element,$attName)
        }
      }
      lappend elements [list $element $attNames $atts]
    }
  }
	return $elements
}
  
proc findDetails {graphtext highlight} {
  global MarkupGui
  
  if [info exists MarkupGui(ttsCmd)] {
    set tmpGraphtext [copyTextWidget $graphtext $graphtext.copy]
    foreach {indx1 indx2} $highlight break
    set PCtag "MARKER EMBED PCend"
    requestTag $PCtag "" $tmpGraphtext "$indx2 $indx2"
    set PCtag "MARKER EMBED PCstart"
    requestTag $PCtag "" $tmpGraphtext "$indx1 $indx1"
    graph2sable $tmpGraphtext
    set text [$tmpGraphtext get 0.0 end]
    destroy $tmpGraphtext
    puts "DB: findDetails: $text"
    # note: may need to limit scope of text in the future. 
    foreach {wave trans} [$MarkupGui(ttsCmd) tts $text] break
    nuke $wave
    array set tmpArray $trans
    # parray tmpArray
    set embed $tmpArray(embed,1)
    set walign $tmpArray(walign,1)
    set palign $tmpArray(palign,1)
    set f0pts $tmpArray(f0,2)
    # find PCstart and PCend
    set PCstart 0; set PCend 0; #safeguard
    foreach elem $embed {
      foreach {time tag} $elem break
      if ![string compare $tag PCstart] {
        set PCstart $time
      }
      if ![string compare $tag PCend] {
        set PCend $time
      }  
    }
    # correct PCstart to fall at beginning of word
    set i 0; set n [llength $walign]; set do 1; set prevB 0
    while {$i < $n & $do} {
      set elem [lindex $walign $i]
      incr i
      foreach {B E word} $elem break
      #puts "DB: comparing $B of word $word with $PCstart "
      if {$B == $PCstart} {
        if {[string compare $word MARK]} { set do 0 } else { set B -1 }
      }
      # note: MARK word always has B = 0 because pron = nil. If PCstart = 0, adjust to word after MARK. 
      if {$B > $PCstart & $prevB < $PCstart & $B <= $PCend } {
        puts "markupGui: findDetails: corrected start point of pitch contour to beginning of $word"
        set PCstart $B
        set do 0
      }
      set prevB $B
    }
    foreach {MarkupGui(highlightBeginTime) MarkupGui(highlightEndTime)} "$PCstart $PCend" break
    if [expr $PCstart > $PCend] {
      puts "error in parsing MARKER tags for default contour: PCstart > PCend"
      set PCstart $PCend
    }
    # retrieve targets and segments between PCstart and PCend
    # also add last target before contour and first one after, only for display purposes
    set f0contour ""
    set do 1
    foreach pt $f0pts {
      foreach {x y} $pt break
      if {$x >= $PCstart & $x <= $PCend} {
        lappend f0contour $x $y
      } elseif {$x < $PCstart} {
        set f0contour "$x $y"
      } elseif {$x > $PCend & $do == 1} {
        lappend f0contour $x $y
        set do 0
      }
    }
    set segs ""
    foreach seg $palign {
      foreach {B E phon} $seg break
      if {$B >= $PCstart & $E <= $PCend} {
        set B [expr $B-$PCstart]
        set E [expr $E-$PCstart]
        lappend segs "$B $E $phon"
      }
    }
    set segs [filter_phn $segs]
    return [list $segs $f0contour]
  } else {
    return "{0 1 nil} {0 0 1 1}"
  }
}

proc defaultDetails {graphtext highlight w} {
  global MarkupGui
  
  puts "DB: highlight: -$highlight-"
  set MarkupGui(default,PITCH,CONTOUR) ""
  set MarkupGui(default,PRON,SEGDUR) ""
  eraseContour 
  $w.book.details.segs load ""
  set tmpGraphtext [copyTextWidget $graphtext $graphtext.tmp]
  applyPopup $tmpGraphtext $highlight $w
  foreach {segs contour} [findDetails $tmpGraphtext $highlight] break
  destroy $tmpGraphtext
  return [list $segs $contour]
  #okPopup $graphtext {$highlight} $w
  #openPopup $graphtext $w
  #$w.book.details select
}
  
proc requestTag {newTag otherTags graphtext highlight} {
  global MarkupGui
  
  # strategy: 
  # if insert-tag modify existing one or apply new one
	# else for each existing tag: 
  # - if partial overlap, place request for each subtag
	# - if coinciding, replace by new tag
	# - in all other cases, repeat recursion for remaining tags 
  # if list empty, apply tag
  foreach {tagName attNames atts} $newTag break
	if $MarkupGui($tagName,isInsert) {
		set otherTag [findSurrInsertTag $graphtext $highlight]
		if [expr [llength $otherTag] > 0] {
			foreach {element ID} [split $otherTag _] break
			foreach {index1 index2} [$graphtext tag ranges $otherTag] break 
      if ![string compare $element $tagName] {
				# replace otherTag by newTag
        $graphtext delete $otherTag.first $otherTag.last
				$graphtext tag delete $index1 $index2
			}
			set highlight "$index1 $index1"
		} 
	  applyTag $newTag $graphtext "$highlight"
	} elseif [expr [llength $otherTags] == 0] {
    applyTag $newTag $graphtext $highlight
  } else {
    set aTag [lindex $otherTags 0]
    set otherTags [lrange $otherTags 1 end]
		set aTagRange [$graphtext tag ranges $aTag]
    foreach {aTag1 aTag2} $aTagRange break
		foreach {aTagName aTagID} [split $aTag _] break
    foreach {newTag1 newTag2} $highlight break
		# if aTag has a non-empty range
		if [expr [llength $aTagRange] > 0] {
      # if existing tag starts somewhere within new tag and ends after it
      set cond1 [expr [$graphtext compare $newTag1 < $aTag1] & \
                [$graphtext compare $aTag1 < $newTag2] & \
                [$graphtext compare $newTag2 < $aTag2]]
      # if existing tag starts somewhere before new tag and ends within it
      set cond2 [expr [$graphtext compare $aTag1 < $newTag1] & \
                [$graphtext compare $newTag1 < $aTag2] & \
                [$graphtext compare $aTag2 < $newTag2]]
      # if existing tag coincides with new tag
      set cond3 [expr [$graphtext compare $aTag1 == $newTag1] & \
                [$graphtext compare $aTag2 == $newTag2] & \
                ![string compare $tagName $aTagName]] 
    } else {
      set cond1 0; set cond2 0; set cond3 0
    }
    if $cond1 {
      requestTag $newTag $otherTags $graphtext "$newTag1 $aTag1"
      requestTag $newTag $otherTags $graphtext "$aTag1 $newTag2"
    } elseif $cond2 {
      requestTag $newTag $otherTags $graphtext "$newTag1 $aTag2"
      requestTag $newTag $otherTags $graphtext "$aTag2 $newTag2"
    } else {
      if $cond3 {
        $graphtext tag delete $aTag
        # no cleanup of attributes since graphtext can be a clone
      }
      requestTag $newTag $otherTags $graphtext $highlight
    }
  }
}
      
  
proc applyTag {tag graphtext highlight} {
  global MarkupGui
  
  foreach {tagName attNames atts} $tag break
	# don't apply atts that are equal to default
	set newAttNames ""
	set newAtts ""
	set len [llength $attNames]
  for {set i 0} {$i < $len} {incr i} {
    set attName [lindex $attNames $i]
    set att [lindex $atts $i]
    if [string compare $MarkupGui(default,$tagName,$attName) $att] {
			lappend newAttNames $attName
			lappend newAtts $att
		}
	}
	set atts $newAtts
	set attNames $newAttNames
	
	if {[llength $attNames] > 0} {
		set ID [generateTagID]
		set tagNameID $tagName\_$ID
		# register new attribute entries in database
		set len [llength $attNames]
    set note "$tagName"
		for {set i 0} {$i < $len} {incr i} {
			set attName [lindex $attNames $i]
			set att [lindex $atts $i]
      set sableAtt [sableValue $tagName $attName $att]
      append note " $attName='$sableAtt'"
			set MarkupGui(attributes,$tagNameID,$attName) $att
		}

		# apply tag to graphtext with correct visual properties
		switch $tagName {
			VOLUME {
				eval "$graphtext tag add $tagNameID $highlight"
			}
      PITCH {
        eval "$graphtext tag add $tagNameID $highlight"
        $graphtext tag configure $tagNameID -font {-slant italic}
      }
      RATE {
        eval "$graphtext tag add $tagNameID $highlight"
        $graphtext tag configure $tagNameID -underline 1
      }
			BREAK {
				set index1 [lindex $highlight 0]
				$graphtext insert $index1 " ... "
				$graphtext tag add $tagNameID $index1 $index1+5c
				$graphtext tag configure $tagNameID -foreground red  
			}
			MARKER {
        set index1 [lindex $highlight 0]
        set i [lsearch $attNames EMBED]
        if {$i > -1} {
          set mark [lindex $att $i]
        } else {
          set mark MARKER
        }
        foreach {dummy what} [split $mark :] break
        if [info exists what] {
          set imageNum [lsearch $MarkupGui(images) $what]
        } else {
          set imageNum -1
        }
        $graphtext insert $index1 "  "
        if {$imageNum > -1} {
          set im [lindex $MarkupGui(images) $imageNum] 
          $graphtext image create $index1+1c -image $im -name $tagNameID
          $graphtext tag add $tagNameID $index1 $index1+3c
        } else {
          set mark " $mark "
          $graphtext insert $index1 "$mark"
          set index2 "$index1+[string length $mark]c"
          $graphtext tag add $tagNameID $index1 $index2
          $graphtext tag configure $tagNameID -background grey 
        }
			}
      SPEAKER {
        set index1 [lindex $highlight 0]
        set i [lsearch $attNames NAME]
        if {$i > -1} {
          set who [lindex $att $i]
        } else {
          set who SPEAKER
        }
        set imageNum [lsearch $MarkupGui(images) $who]
        $graphtext insert $index1 "  "
        if {$imageNum > -1} {
          set index1 [$graphtext index $index1+1c]
          set im [lindex $MarkupGui(images) $imageNum] 
          $graphtext image create $index1 -image $im -name $tagNameID
          $graphtext tag add $tagNameID $index1 $index1+1c
        } else {
          $graphtext insert $index1 "$who"
          set index2 "$index1+[string length $who]c"
          $graphtext tag add $tagNameID $index1 $index2
          $graphtext tag configure $tagNameID -background grey 
        }
      }
      AUDIO {
        set index1 [lindex $highlight 0]
        set imageNum [lsearch $MarkupGui(images) "audio2"]
        if {$imageNum > -1} {
          $graphtext insert $index1 "  "
          set im [lindex $MarkupGui(images) $imageNum] 
          $graphtext image create $index1+1c -image $im -name $tagNameID
          $graphtext tag add $tagNameID $index1 $index1+3c
        } else {
          set who " $MarkupGui(popup,AUDIO,SRC) "
          $graphtext insert $index1 "$who"
          set index2 "$index1+[string length "$who"]c"
          $graphtext tag add $tagNameID $index1 $index2
          $graphtext tag configure $tagNameID -background grey 
        }
      }
      SAYAS -
      PRON {
        eval "$graphtext tag add $tagNameID $highlight"
        $graphtext tag configure $tagNameID -foreground red
      }
      default {
        puts "don't know how to visualize $tagName"
      }
		}
    StickyNote attach $graphtext $note $tagNameID
  }
}  
  
proc generateTagID {} {
  global MarkupGui
  
  if [info exists MarkupGui(ID)] {
    incr MarkupGui(ID)
  } else {
    set MarkupGui(ID) 1
  }
  return $MarkupGui(ID)
}

# input text widget will be sable-ized. Calling procedure is responsible for
# making a copy first
proc graph2sable {sabletext} {
  global MarkupGui

  set tags [findSableTags $sabletext]
  # sort the tags to maintain proper embedding on the sable tags
  set tags [longestFirst $sabletext $tags]
	# now replace text widget tags by sable tags
  foreach tag $tags {
    set highlight [$sabletext tag ranges $tag]
    foreach {element ID} [split $tag _] break
    #puts "graph2sable $element $highlight" ; vwait pauseVar
		set sableTag "<$element"
		foreach attName $MarkupGui($element,attNames) {
			if [info exists MarkupGui(attributes,$element\_$ID,$attName)] {
				set att $MarkupGui(attributes,$element\_$ID,$attName)
        set att [sableValue $element $attName $att]
				append sableTag " $attName='$att'" 
			} 
		}
    if $MarkupGui($element,isInsert) {
      append sableTag "/>"
			eval "$sabletext delete $highlight"
      $sabletext insert [lindex $highlight 0] $sableTag
      $sabletext tag delete $tag
		} else {
			append sableTag ">"
      $sabletext insert [lindex $highlight 1] "</$element>"
      $sabletext insert [lindex $highlight 0] $sableTag
      $sabletext tag delete $tag
    }
  }
  # surround by <SABLE> tags
  $sabletext insert end-2c </SABLE>
  $sabletext insert 1.0 "<SABLE> "
}

# translate attribute from graphtext to sabletext
# e.g. integer slider value is translated to +/- percent value
# e.g.2 list box value is translated into correct sable abbreviation
proc sableValue {element attName att} {
  global MarkupGui
  
  set key "$element\_$attName"
  switch $key {
    PITCH_BASE   -
    PITCH_RANGE  -
    VOLUME_LEVEL -
    RATE_SPEED  {
      set relVal [expr $att - 100]
      if {$relVal < 0} {set sign ""} else {set sign +}
      return "$sign$relVal\%"
    }
    PRON_IPA {
      # replace & by &amp; as is required by XML standard
      regsub -all "&" $att "&amp;" att
      return $att
    }
    PRON_SEGDUR {
      # change from seglist {{0 100 A} {100 150 p} ..} into "A 100; p 50; .."
      regsub -all "&" $att "&amp;" att
      set s ""
      foreach seg $att {
        foreach {B E phon} $seg break
        set dur [expr round($E - $B)]
        append s "$phon $dur;"
      }
      return $s
    }
    BREAK_LEVEL {
      set breakType [lindex $MarkupGui(breakTypes) $att]
      return [lindex $breakType 1]
    }
    AUDIO_SRC {
			if ![regexp {.*\.wav} $att] {
        return "$att.wav"
			} else {
				return $att
			}
    }
    VOLUME_CONTOUR -
    PITCH_CONTOUR {
      set contour ""
      foreach {x y} $att {
        append contour "$x $y;"
      }
      return $contour
    }
    MARKER_EMBED {
      foreach {type what} [split $att :] break
      if {$type == "emo" & $what != ""} {
        if [array exists tmpArray] {unset tmpArray}
        foreach elem $MarkupGui(baldiEmotions) {
          foreach {name desc} $elem break
          set tmpArray($name) $desc
        }
        if [info exists tmpArray($what)] {
          return "_emo:n1.0:$tmpArray($what)"
        } 
      } elseif {$type == "gest" & $what != ""} {
        return "_ges:$what,100"
      } else {
        return $att
      }
    }
    BREAK_MSEC -
    default {
      return $att
    }
  }
}

# translate attribute from sabletext to graphtext
# this is the reverse of sableValue
proc graphValue {element attName att} {
  global MarkupGui
  
  set key "$element\_$attName"
  switch $key {
    PITCH_BASE   -
    PITCH_RANGE  -
    VOLUME_LEVEL -
    RATE_SPEED  {
			regsub -all {\+|%} $att {} att
			if [regexp -- {-?[0-9]*.?[0-9]+} $att] {
        return [expr $att + 100]
			} else {
				return 100
			}
    }
    PRON_SEGDUR {
      # change from "A 100; p 50; .." into seglist {{0 100 A} {100 150 p} ..}
      # note that XML parser substitutes &amp; to &
      puts "DB: $att"
      set seglist ""
      set B 0
      foreach seg [split $att ";"] {
        foreach {phon dur} $seg break
        if {$phon == "" | ![regexp {[0-9]*.[0-9]+} $dur]} { return $seglist }
        set E [expr $B + $dur]
        lappend seglist [list $B $E $phon]
      }
      return $seglist
    }
    BREAK_LEVEL {
      set indx 0
      if {$att != ""} {
  			set indx [lsearch $MarkupGui(breakTypes) $att]
      }
			return $indx
    }
    AUDIO_SRC {
			if [regexp {.*\.wav} $att] {
				return [file rootname $att]
			} else {
				return $att
			}
    }
    VOLUME_CONTOUR -
    PITCH_CONTOUR {
      regsub -all ";" $att { } att
      regsub -all {[^0-9. ]} $att {} att
      if {[regexp {\.[0-9]+\.} $att tmp] | fmod([llength $att],2) > 0} {
        return ""
			} else {
				return $att
			}
    }
    MARKER_EMBED {
			if [regexp {_ges:(.*),100} $att dummy name] {
				return "ges:$name"
			}
			foreach elem $MarkupGui(baldiEmotions) {
				foreach {name desc} $elem break
				if {$att == "_emo:n1.0:$desc"} { return emo:$name }
			}
			return $att
    }
    BREAK_MSEC -
    default {
      return $att
    }
  }
}

proc normalizeContour {oldContour} {
  global MarkupGui
  
  set contour ""
  set B $MarkupGui(highlightBeginTime)
  set E $MarkupGui(highlightEndTime)
  set dur [expr $E-$B]
  if {$dur == 0 } { return $contour }
  foreach {x y} $oldContour {
    if {$x >= $B & $x <= $E} {
      set x [expr ($x-$B)/$dur]
      set x [format "%.4f" $x] ;# precision makes sable less readable, but is necessary for large dur
      set y [format "%.2f" $y]
      lappend contour $x $y
    }
  }
  return $contour
}

proc normalizeSegs {oldsegs} {
  set segs ""
  foreach seg $oldsegs {
    foreach {B E phon} $seg break
    set B [format "%.2f" $B]
    set E [format "%.2f" $E]
    lappend segs [list $B $E $phon]
  }
  return $segs
}

proc findSableTags {graphtext args} {
  set tags [eval "$graphtext tag names $args"]
  set sableTags ""
  foreach tag $tags {
    if [regexp {^.*_[0-9]+$} $tag] {
      lappend sableTags $tag
    }
  }
  return $sableTags
}
  
# only the text, tags and images are copied at the moment
proc copyTextWidget {graphtext name} {
  if ![winfo exists $name] {
	  text $name 
	} else {
		$name delete 1.0 end
	}
  $name insert 1.0 [$graphtext get 1.0 end]
  set images [$graphtext image names]
  foreach image $images {
    set indx [$graphtext index $image]
    set im [$graphtext image configure $indx -image]
    set im [lindex $im 4]
    if {$indx != ""} {
      eval "$name image create $indx -image $im"
    }
  }
  set tagNames [$graphtext tag names]
  foreach tagName $tagNames {
    set highlight [$graphtext tag ranges $tagName]
    if {$highlight != ""} {
		  eval "$name tag add $tagName $highlight"
			## for debugging only:
			if 0 { 
				set tagOptions [$graphtext tag configure $tagName] 
				foreach tagOption $tagOptions {
					set optionName [lindex $tagOption 0]
					set optionVal [lindex $tagOption 4]
					if {$optionVal != {}} {
						$name tag configure $tagName $optionName $optionVal
					}
				}
			}
		}
  }
  return $name
}

# the text widget doesn't combine the font specs from different tags
# so here we lay all the tags down starting with the longest and propagate font changes
proc propagateFontChanges {graphtext} {
  global MarkupGui
  
  set tags [findSableTags $graphtext]
  set tags [longestFirst $graphtext $tags]
  set len [llength $tags]
  set doneTags ""
  for {set i 0} {$i < $len} {incr i} {
    set tag [lindex $tags $i]
    foreach {element ID} [split $tag _] break
    if {$element == "PITCH" | $element == "VOLUME"} {
      array set fontArray [font actual [$graphtext cget -font]] 
      set highlight [$graphtext tag ranges $tag]
      set surTags [findSurroundingTags $graphtext $highlight]
      set surTags [intersect $surTags $doneTags]
      set surTags [longestFirst $graphtext $surTags] 
      foreach oldTag $surTags {
        if {$oldTag != $tag} {
          set oldfont [$graphtext tag cget $oldTag -font]
          array set tmpArray $oldfont
          foreach name [array names tmpArray] {
            set fontArray($name) $tmpArray($name)
          }
        }
      }
      # changing font family of a tag causes weird behavior in text widget...
      if [info exists fontArray(-family)] { unset fontArray(-family) }
      switch $element {
        VOLUME {
          set sign +
          if [info exists MarkupGui(attributes,$tag,LEVEL)] {
            set level $MarkupGui(attributes,$tag,LEVEL)
            if {$level < 100} { set sign - }
          } 
          if [info exists fontArray(-size)] {
            incr fontArray(-size) "$sign $MarkupGui(fontSizeIncrease)"
            if {$fontArray(-size) < $MarkupGui(minFontSize)} {
              set fontArray(-size) $MarkupGui(minFontSize)}
          } else {
            set fontArray(-size) $MarkupGui(largeFontSize)
          }
        }
        PITCH {
          set fontArray(-slant) italic
        }
      }
      set newfont [array get fontArray]
      $graphtext tag configure $tag -font $newfont
      $graphtext tag raise $tag
      lappend doneTags $tag
    }
  }
}      
  

proc longestFirst {graphtext tags} {
  # rank tags from longest to shortest. 
  set tmpTags ""
  foreach tag $tags {
    if {[$graphtext tag ranges $tag] != ""} {
      lappend tmpTags "$tag [string length [$graphtext get $tag.first $tag.last]]"
    }
  }
  set tmpTags [lsort -integer -decreasing -index 1 $tmpTags]
  set tags ""
  foreach tag $tmpTags {
    lappend tags [lindex $tag 0]
  }
  return $tags
}

proc intersect {list1 list2} {
  # find shortes list
  if [expr [llength $list2] < [llength $list1]] {
    set list3 $list1
    set list1 $list2
    set list2 $list3
  }
  set newlist ""
  foreach elem $list1 {
    if [expr [lsearch -exact $list2 $elem] > -1] {
      lappend newlist $elem
    }
  }
  return $newlist
}

proc initFakeSel { w } {
	$w tag configure selFake \
 	  -foreground SystemHighlightText -background SystemHighlight
}

proc fakeSelect { w action args} { 
 catch {$w tag remove selFake selFake.first selFake.last }
 #catch {$w tag remove sel sel.first sel.last }
 if { $action == "destroy" } { return }
 if [expr [llength $args] > 0] {
	 initFakeSel $w
	 eval "$w tag add selFake [lindex $args 0]"
 }
}

proc filter_phn {segs} {
  set filtphn ""
  set prevp ""
  foreach line $segs {
    foreach {b e p} $line {}
    regsub {.pau} $p {pau} p
    if [regsub {(.*)c} $prevp {\1} pp] {
      if [regexp "^$pp\h?$" $p] {
        set b $prevb
        set p $pp
      } else {
        lappend filtphn [list $prevb $preve $pp]
      }
    } else {
      if {$prevp != ""} {
        lappend filtphn [list $prevb $preve $prevp]
      }
    }
    set prevb $b
    set preve $e
    set prevp $p
  }
  if [info exists prevb] {
    set filtphn [lappend filtphn [list $prevb $preve $prevp]]
  }
  return $filtphn
}

proc initializeMarkupGui {} {
  global MarkupGui
	# global variable MarkupGui contains the following fields

	# default values of the graphical elements
	set MarkupGui(default,SABLE,) ""
	set MarkupGui(default,PITCH,BASE) 100
  set MarkupGui(default,PITCH,RANGE) 100
  set MarkupGui(default,PITCH,TOBI) NONE
  set MarkupGui(default,PITCH,CONTOUR) ""
  set MarkupGui(default,VOLUME,CONTOUR) ""
	set MarkupGui(default,RATE,SPEED) 100
	set MarkupGui(default,VOLUME,LEVEL) 100
	set MarkupGui(default,listbox,selection) 0
  set MarkupGui(default,SAYAS,MODE) ""
  set MarkupGui(default,SAYAS,MODETYPE) ""
  set MarkupGui(default,PRON,SUB) ""
  set MarkupGui(default,PRON,IPA) ""
  set MarkupGui(default,PRON,SEGDUR) ""
	set MarkupGui(default,BREAK,LEVEL) 0  ;# default is automatic break assignment
	set MarkupGui(default,BREAK,MSEC) 0 ;# default is 0 msec pause between words
  set MarkupGui(default,MARKER,EMBED) ""
  set MarkupGui(default,SPEAKER,NAME) ""
  set MarkupGui(default,AUDIO,SRC) ""

	# elementNames and possible attribute names for each element are
	# derived from default values
	set names [array names MarkupGui "default,*,*"]
	set MarkupGui(elementNames) ""
	foreach name $names {
		foreach {default element att} [split $name ,] break
		if [expr [lsearch $MarkupGui(elementNames) $element] < 0] {
			lappend MarkupGui(elementNames) $element
			set MarkupGui($element,attNames) ""
		}
		if [expr [lsearch $MarkupGui($element,attNames) $att] < 0] {
			lappend MarkupGui($element,attNames) $att
		}
	}

	# insert type for each element
	set insertElements "BREAK MARKER AUDIO SPEAKER BALDI SOUNDS"
	foreach elem $MarkupGui(elementNames) {
		set MarkupGui($elem,isInsert) 0
	}
	foreach elem $insertElements {
		set MarkupGui($elem,isInsert) 1
	}

	# other fields
	catch {font create largeFont -family Helvetica -size 12 -weight bold}
  set MarkupGui(largeFont) largeFont
  set MarkupGui(minFontSize) 5
	set MarkupGui(images) ""
  set MarkupGui(breakTypes) "{Automatic {}} {{No Break} NONE} {{Phrase Break} B} \
	  {{Sentence Break} LARGE} {Pause... {}}"
  set MarkupGui(sayasCommands) "Normal Pronounce Spell Digits Amount Syllabify"
  set MarkupGui(baldiEmotions) { \
      {angry     ang=1.0/dis=0.0/fea=0.0/hap=0.0/sad=0.0/sur=0.0/non=0.0}\
      {disgusted ang=0.0/dis=1.0/fea=0.0/hap=0.0/sad=0.0/sur=0.0/non=0.0}\
      {fearful   ang=0.0/dis=0.0/fea=1.0/hap=0.0/sad=0.0/sur=0.0/non=0.0}\
      {happy     ang=0.0/dis=0.0/fea=0.0/hap=1.0/sad=0.0/sur=0.0/non=0.0}\
      {sad       ang=0.0/dis=0.0/fea=0.0/hap=0.0/sad=1.0/sur=0.0/non=0.0}\
      {surprised ang=0.0/dis=0.0/fea=0.0/hap=0.0/sad=0.0/sur=1.0/non=0.0}\
      {neutral   ang=0.0/dis=0.0/fea=0.0/hap=0.0/sad=0.0/sur=0.0/non=1.0}}
  set MarkupGui(baldiGestures) {armsdown armsbehind display1 emph1 holdon \
     moi referleft1 referleft2 referright1 referright2 shrug}
  set emphasisPresets {{none} \
     {moderate PITCH,TOBI H* VOLUME,LEVEL 120} \
     {strong PITCH,TOBI H* PITCH,RANGE 120 VOLUME,LEVEL 150 RATE,SPEED 80 VOLUME,LEVEL 150}}
  
  # construct MarkupGui(emphasisPresets) array from standard or custom file if found
  set MarkupGui(emphasisPresets) "" 
  foreach preset $emphasisPresets {
    set tmpList ""
    foreach key {PITCH,BASE PITCH,RANGE PITCH,TOBI VOLUME,LEVEL RATE,SPEED} {
      lappend tmpList $key $MarkupGui(default,$key)
    }
    array set tmpArray $tmpList
    set name [lindex $preset 0]
    foreach {key value} [lrange $preset 1 end] {
      set tmpArray($key) $value
    }
    lappend MarkupGui(emphasisPresets) "$name [array get tmpArray]" 
  }
  if [file exists c:/.markupGui] {
    set fp [open c:/.markupGui] 
    set text [read $fp]
    close $fp
    array set tmpArray $text
    if [info exists tmpArray(emphasisPresets)] {
      set MarkupGui(emphasisPresets) $tmpArray(emphasisPresets)
    }
  }

	set MarkupGui(largeFontSize) 18
	set MarkupGui(fontSizeIncrease) 4
	
	# remembered values of the graphical elements
	# are set in proc rememberElementValues and have form MarkupGui(remember,X)

	# current values of the graphical elements
	# are set initially in proc setElementValues and have the form MarkupGui(popup,X)
	# these variables are modified directly by the graphical elements in the popup
}

proc saveSettings {} {
  global MarkupGui
  
  # only emphasisPresets at the moment
  set fp [open c:/.markupGui w]
  puts $fp "emphasisPresets {$MarkupGui(emphasisPresets)}"
  close $fp
}
  
proc markupSetTTSCommand {ttsCmd} {
  global MarkupGui
  
  set MarkupGui(ttsCmd) $ttsCmd
}

proc markupDeleteImages {} {
  global MarkupGui
 
  eval "image delete $MarkupGui(images)"
}

proc markupCreateImages {args} {
  global MarkupGui
  
  if [info exists MarkupGui(images)] {
    markupDeleteImages
  }
  set MarkupGui(images) ""
  foreach dir $args {
    foreach file [glob $dir/*.gif] {
      set name [file tail [file rootname $file]]
      if [catch {
        set im [image create photo $name -file $file]
        lappend MarkupGui(images) $im
      } msg] { puts $msg }
    }
  }
}

proc markupFindSounds {args} {
  global MarkupGui
  
  set MarkupGui(sounds) ""
  foreach dir $args {
    foreach file [glob $dir/*.wav] {
      set name [file tail [file rootname $file]]
      lappend MarkupGui(sounds) $name
    }
  }
}
  